import javax.swing.*;  
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

/** Panel for holding any images.
 */
class ImagePanel extends JPanel implements MouseListener, MouseMotionListener{

	private BufferedImage bimg; // The image currently on display
	private BufferedImage prevImg; // previous image.
	private int state=0;		// action state of the panel.
	// objects drawn on the pannel.
	private java.util.Vector objs=new java.util.Vector();
	private int selectionX=0, selectionY=0; // selection coordinates.
	private ICOperation operation;	// Operation object for some state.
	
	// State constants.
	
	/** Regular state of the panel. */
	public static final int POINTER=0;
	/** State for selecting part of the image. */
	public static final int SELECT=1;
	/** Miscellaneous state. */
	public static final int MISC=2;
	/** Scissoring state. */
	public static final int SCISSOR=3;

	/** Constructor: */
	public ImagePanel(){
		addMouseListener(this);
		addMouseMotionListener(this);
		addKeyListener(KeyBoardManager.getKeyBoardManager());
	}
	
	/** Constructor:
	 * @param bimg Image to put in the panel.
	 */
	public ImagePanel(BufferedImage bimg){
		addMouseListener(this);
		addMouseMotionListener(this);
		addKeyListener(KeyBoardManager.getKeyBoardManager());
		setImage(bimg);
	}
	
	/** Constructor:
	 * @param filename Path to the image file.
	 */
	public ImagePanel(String filename){
		addMouseListener(this);
		addMouseMotionListener(this);
		addKeyListener(KeyBoardManager.getKeyBoardManager());
		setImage(filename);
	}
	
	/** Puts image in the file into this panel.
	 * @param filename File path of the image.
	 */
	public boolean setImage(String fileName){
		BufferedImage b=ImageLoader.loadImage(fileName);
		if(b==null){ bimg=null;  return false; }
		setImage(b);
		return true;
	}

	/** Puts an image into this panel.
	 * @param bimg Image.
	 */
	public void setImage(BufferedImage bimg){
		if(bimg==null)return;
		prevImg=this.bimg;
		this.bimg=bimg;
		Dimension size=new Dimension(bimg.getWidth(this),bimg.getHeight(this));
		setMinimumSize(size); setPreferredSize(size);
		repaint();
	}

	/** Paints image 'b' on top of the image in this panel.
	 * @param b Image to paint.
	 * @param x X coordinate to place the image at.
	 * @param y Y coordinate to place the image at.
	 */
	public void paintImage(BufferedImage b, int x, int y){ 
		if(bimg==null)return;
		prevImg=copyImage();
		Graphics2D g=bimg.createGraphics(); 
		g.drawImage(b,null,x,y); 
		repaint();
	}

	/** Returns the image in this panel.
	 * @return The image in this panel.
	 */
	public BufferedImage getImage(){ return bimg; }

	/** Returns a copy of this image.
	 * @return Copy of the image in this panel.
	 */
	public BufferedImage copyImage(){ 
		if(bimg==null)return null;
		BufferedImage b=new BufferedImage(bimg.getWidth(), bimg.getHeight(), BufferedImage.TYPE_INT_ARGB);
		Graphics2D g=b.createGraphics();
		g.drawImage(bimg,null,0,0);
		return b;
	}

	/** Returns image width.
	 * @return Image width.
	 */
	public int getImageWidth(){ if(bimg==null)return -1; return bimg.getWidth(this); }
	/** Returns image height.
	 * @return Image height.
	 */
	public int getImageHeight(){ if(bimg==null)return -1; return bimg.getHeight(this); }

	/** Paints this component.
	 * @param g Graphics object to paint on.
	 */
	public void paintComponent(Graphics g){
		super.paintComponent(g);
		if(bimg!=null)g.drawImage(bimg,0,0,this);
		
		if(operation!=null)
			operation.beforeDrawing();
		for(int i=0; i<objs.size(); i++)
			((Shape)objs.elementAt(i)).paint(g);
	}
	
	/** Event handler for mouse presses.
	 * @param e Event object describing the mouse event.
	 */
	public void mousePressed(MouseEvent e) {
		switch(state){
			case POINTER:
				if(objs.size()==1 && objs.elementAt(0) instanceof SelectionShape){
					// this is a selection
					SelectionShape s=(SelectionShape)objs.elementAt(0);
					if(s.hitTest(e.getX(), e.getY())){
						// mouse within selection, so don't deselect.
						
					}else{
						objs.clear();
						selectionX=selectionY=0;
					}
				}
				repaint();	
				break;
			case SELECT:
				if(operation.mousePressed(e)) repaint();
				break;
		}
	}
	/** Event handler for mouse releases.
	 * @param e Event object describing the mouse event.
	 */
	public void mouseReleased(MouseEvent e) {
		switch(state){
			case SELECT:
				if(operation.mouseReleased(e)) repaint();
				break;
		}
	}
	/** Event handler for mouse enterings.
	 * @param e Event object describing the mouse event.
	 */
	public void mouseEntered(MouseEvent e) {
		switch(state){
			case SELECT:
				if(operation.mouseEntered(e)) repaint();
				break;
			case SCISSOR:
				if(operation.mouseEntered(e)) repaint();
				break;
		}
	}
	/** Event handler for mouse exits.
	 * @param e Event object describing the mouse event.
	 */
	public void mouseExited(MouseEvent e) {
		switch(state){
			case POINTER:
				setCursor(Laf.getCursor("default"));
				break;
				
			case SELECT:
				if(operation.mouseExited(e)) repaint();
				break;
				
			case SCISSOR:
				if(operation.mouseExited(e)) repaint();
				break;
		}
	}
		
	/** Event handler for mouse clicks.
	 * @param e Event object describing the mouse event.
	 */
	public void mouseClicked(MouseEvent e) {
		switch(state){
			case POINTER:
				if(objs.size()==1 && objs.elementAt(0) instanceof RectSelection){
					// this is a selection
					SelectionShape s=(SelectionShape)objs.elementAt(0);
					if(s.hitTest(e.getX(), e.getY())){
						// mouse within selection, so don't deselect.
						
					}else{
						objs.clear();
						selectionX=selectionY=0;
					}
				}
				repaint();
				break;	
				
			case SCISSOR:
				if(operation.mouseClicked(e)) repaint();
				break;
		}
	}

	/** Event handler for the mouse move event.
	 * @param e Object describing the event.
	 */
	public void mouseMoved(MouseEvent e) {
		switch(state){
			case SCISSOR:
				if(operation.mouseMoved(e)) repaint();
				break;
		}
	}

	/** Event handler for the mouse drag event.
	 * @param e Object describing the event.
	 */
	public void mouseDragged(MouseEvent e) {
		switch(state){
			case SELECT:
				if(operation.mouseDragged(e)) repaint();
				break;	
				
			case SCISSOR:
				if(operation.mouseDragged(e)) repaint();
				break;
		}
		repaint();
	}
	
	/**
	 * Sets the state of the panel.
	 * @param state One of the state constants.
	 */
	public void setState(int state){
		operation=null;
		switch (state){
			case POINTER:
				if(objs.size()!=1 || !(objs.elementAt(0) instanceof SelectionShape))
					objs.clear();
				break;
			case SELECT:
				objs.clear();
				operation=new SelectOp(objs, this);
				break;
			case SCISSOR:
				objs.clear();
				operation=new Scissor(getImage(), objs, this);
				break;
		}
		this.state=state;
		repaint();
	}
	
	/**
	 * Returns the state.
	 * @return State of the panel.
	 */
	public int getState(){
		return state;
	}
	
	/**
	 * Pastes an image into this image panel.
	 * @param pimg Image to paste.
	 */
	public void pasteImage(BufferedImage pimg){
		if(getImage()==null) return;
		BufferedImage im=pimg;
		// x & y positions of the pasted image.
		int xPos=(getImageWidth()-pimg.getWidth())/2, yPos=(getImageHeight()-pimg.getHeight())/2;
		
		if(objs.size()==1 && objs.elementAt(0) instanceof SelectionShape){
			// this is a selection
			SelectionShape s=(SelectionShape)objs.elementAt(0);
			// recalculate the position of pasted image with relation to selection image.
			xPos=(s.getWidth()-pimg.getWidth())/2+getSelectionX();
			yPos=(s.getHeight()-pimg.getHeight())/2+getSelectionY();
			// new image for cutting out the selected portion of the pasted image.
			im=new BufferedImage(getImageWidth(), getImageHeight(), BufferedImage.TYPE_INT_ARGB);
			im.createGraphics().drawImage(pimg, xPos, yPos, null);
			// cut out the selected portion of the pasted image.
			im=s.getSelection(im);
			xPos=getSelectionX();
			yPos=getSelectionY();
		}
		paintImage(im, xPos, yPos);
	}
	
	/**
	 * Returns the image of the selected area.
	 * @return Image for the selected area.
	 */
	public BufferedImage getSelection(){
		if(objs.size()==1 && objs.elementAt(0) instanceof SelectionShape){
			// this is a selection
			SelectionShape s=(SelectionShape)objs.elementAt(0);
			return s.getSelection(getImage());
		}
		return null;
	}
	
	/**
	 * Returns x position of the selection.
	 * @return X position of the selection.
	 */
	public int getSelectionX(){
		return selectionX;
	}
	
	/**
	 * Returns the y position of the selection.
	 * @return Y position of the selection.
	 */
	public int getSelectionY(){
		return selectionY;
	}
	
	/**
	 * Set selection's x position.
	 * @param x New x position.
	 */
	public void setSelectionX(int x){
		selectionX=x;
	}

	/**
	 * Set selection's y position.
	 * @param y New y position.
	 */
	public void setSelectionY(int y){
		selectionY=y;
	}
	
	/**
	 * Puts the previous image into the panel.
	 */
	public void setPrevState(){
		setImage(prevImg);
	}
	
	/**
	 * Returns the previous image.
	 * @return Previous image.
	 */
	protected BufferedImage getPrevImage(){
		return prevImg;
	}
	
	/**
	 * Does clean up before thw indow with this panel is closed.
	 */
	public void cleanUp(){}
	
	/**
	 * Finalizes some operation this image panel was doing.
	 * @return true if finilization was completed, false otherwise.
	 */
	public boolean finilizeOperation(){ return true;}
}
