import java.awt.Color;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.util.Vector;


/**
 * Shape for a general selection.
 */
public class SelectionShape extends Shape {
	/**
	 * Constructor:
	 */
	public SelectionShape(){
		super();
	}
	
	/**
	 * Constructor:
	 * @param c Color of the shape.
	 */
	public SelectionShape(Color c){
		super(c);
	}
	
	/**
	 * Constructor:
	 * @param v Path of the selection.
	 */
	public SelectionShape(Vector v){
		super(v);
	}
	
	/**
	 * Constructor:
	 * @param v Path of the selection.
	 * @param c Color of the selection path.
	 */
	public SelectionShape(Vector v, Color c){
		super(v, c);
	}

	/**
	 * Returns the x position of the selection.
	 * @return X position.
	 */
	public int getX(){
		XYPair xy;
		// find min, max.
		int minX=-1;
		for(int i=0; i<pos.size(); i++){
			xy=(XYPair)pos.elementAt(i);
			if(xy.x<minX || minX==-1) minX=xy.x;
		}
		return minX;
	}

	/**
	 * Returns the y position of the selection.
	 * @return Y position.
	 */
	public int getY(){
		XYPair xy;
		// find min, max.
		int minY=-1;
		for(int i=0; i<pos.size(); i++){
			xy=(XYPair)pos.elementAt(i);
			if(xy.y<minY || minY==-1) minY=xy.y;
		}
		return minY;
	}

	/**
	 * Returns the width of the selection.
	 * @return Width of the selection.
	 */
	public int getWidth(){
		XYPair xy;
		int minX=-1, maxX=0;
		
		// find min, max.
		for(int i=0; i<pos.size(); i++){
			xy=(XYPair)pos.elementAt(i);
			if(xy.x>maxX) maxX=xy.x;
			if(xy.x<minX || minX==-1) minX=xy.x;
		}
		return maxX-minX+1;
	}

	/**
	 * Returns the height of the selection.
	 * @return Height of the selection.
	 */
	public int getHeight(){
		XYPair xy;
		int minY=-1, maxY=0;
		
		// find min, max.
		for(int i=0; i<pos.size(); i++){
			xy=(XYPair)pos.elementAt(i);
			if(xy.y>maxY) maxY=xy.y;
			if(xy.y<minY || minY==-1) minY=xy.y;
		}
		return maxY-minY+1;
	}
	
	/**
	 * tests if a point is within the shape.
	 * @param x X coordinate.
	 * @param y Y coordinate.
	 * @return true if the point is withint the shape, false otherwise.
	 */
	public boolean hitTest(int x, int y){
		XYPair xy;
		int minX=-1, maxX=0, minY=-1, maxY=0;
		
		// find min, max.
		for(int i=0; i<pos.size(); i++){
			xy=(XYPair)pos.elementAt(i);
			if(xy.y==y && xy.x==x)
				return true;
				
			if(xy.y>maxY) maxY=xy.y;
			if(xy.y<minY || minY==-1) minY=xy.y;
			if(xy.x>maxX) maxX=xy.x;
			if(xy.x<minX || minX==-1) minX=xy.x;
		}
		if(x<minX || x>maxX || y<minY || y>maxY)
			return false;
			
		byte [][] visited=new byte[maxY-minY+3][maxX-minX+3];
		// fill in pixels along the selection path.
		for(int i=0; i<pos.size(); i++){
			visited[((XYPair)pos.get(i)).y-minY+1][((XYPair)pos.get(i)).x-minX+1]=2;
		}
		
		// that's why selection and visited are bigger than the image by 1 row
		// because this last row is guaranteed not to have any pixels on the selection line.
		visited[0][0]=1;
		x-=minX-1;
		y-=minY-1;
		Vector unvisited=new Vector();
		// Go through every pixel and see if it touches any neighbor that was visited before.
		for(int i=0; i<visited.length; i++){
			for(int j=0; j<visited[0].length; j++){
				// if it's visited or on the selection path skip it
				if(visited[i][j]>0) continue;
				
				// Check if any of the neighbors are visited.
				if((visited[i][j]=checkHeighbors(visited, j, i))==0)
					// No visited neighbors so add to to the list of unvisited pixels.
					unvisited.add(new XYPair(j, i));
			}
		}
		int prevSize=unvisited.size()+1;
		
		// Go through the unvisited pixels checking if some of them now
		// touch a visited neighbor, do this until you can't make the rest
		// of the pixels visited.
		while(prevSize!=unvisited.size()){
			prevSize=unvisited.size();
			for(int i=0; i<unvisited.size(); i++){
				xy=(XYPair)unvisited.elementAt(i);
				// if it's on the selection path skip it.
				if(visited[((XYPair)unvisited.get(i)).y][((XYPair)unvisited.get(i)).x]==2)
					continue;
				
				// Check if any of the neighbors are now visited.
				if((visited[xy.y][xy.x]=checkHeighbors(visited, xy.x, xy.y))==1){
					// it was visited so remove it from the list.
					unvisited.remove(i);
					i--;
				}
			}
		}

		if(visited[y][x]==1)
			return false;
		return true;
	}
	
	/** Checks whether any of the horizontal or vertical neighbors are visited.
	 * @param visited 2D boolean array that tells which pixels have been visited.
	 * @param x X coordinate of the pixel to check.
	 * @param y Y coordinate of the pixel to check.
	 * @return true if at least 1 of the neighbors is visited, false otherwise.
	 */
	private byte checkHeighbors(byte [][] visited, int x, int y){
		if(y>0 && visited[y-1][x]==1)
			return 1;
		else if(y<visited.length-1 && visited[y+1][x]==1)
			return 1;
		else if(x>0 && visited[y][x-1]==1)
			return 1;
		else if(x<visited[0].length-1 && visited[y][x+1]==1)
			return 1;
		return 0;
	}
	
	/**
	 * Paints the shape.
	 * @param g Graphics object to paint on.
	 */
	public void paint(Graphics g){
		Color c=g.getColor();
		g.setColor(color);
		// go through all the points on the path.
		for(int i=0; i<pos.size(); i+=4){
			// draw 2 points in a row and skip next 2.
			g.fillRect(((XYPair)pos.get(i)).x, ((XYPair)pos.get(i)).y, 1, 1);
			if(i+1<pos.size())
				g.fillRect(((XYPair)pos.get(i+1)).x, ((XYPair)pos.get(i+1)).y, 1, 1);
		}
		g.setColor(c);
	}
	
	/**
	 * Returns the area of the image within this shape.
	 * @param im Image to get the selection for.
	 * @return Area of the image within this shape.
	 */
	public BufferedImage getSelection(BufferedImage im){
		XYPair xy;
		int minX=-1, maxX=0, minY=-1, maxY=0;
		
		if(im==null) return null;
		
		// find min, max.
		for(int i=0; i<pos.size(); i++){
			xy=(XYPair)pos.elementAt(i);
			if(xy.y>maxY) maxY=xy.y;
			if(xy.y<minY || minY==-1) minY=xy.y;
			if(xy.x>maxX) maxX=xy.x;
			if(xy.x<minX || minX==-1) minX=xy.x;
		}
			
		byte [][] visited=new byte[maxY-minY+3][maxX-minX+3];
		// fill in pixels along the selection path.
		for(int i=0; i<pos.size(); i++){
			visited[((XYPair)pos.get(i)).y-minY+1][((XYPair)pos.get(i)).x-minX+1]=2;
		}
		
		// that's why selection and visited are bigger than the image by 1 row
		// because this last row is guaranteed not to have any pixels on the selection line.
		visited[0][0]=1;
		Vector unvisited=new Vector();
		// Go through every pixel and see if it touches any neighbor that was visited before.
		for(int i=0; i<visited.length; i++){
			for(int j=0; j<visited[0].length; j++){
				// if it's visited or on the selection path skip it
				if(visited[i][j]>0) continue;
				
				// Check if any of the neighbors are visited.
				if((visited[i][j]=checkHeighbors(visited, j, i))==0)
					// No visited neighbors so add to to the list of unvisited pixels.
					unvisited.add(new XYPair(j, i));
			}
		}
		int prevSize=unvisited.size()+1;
		
		// Go through the unvisited pixels checking if some of them now
		// touch a visited neighbor, do this until you can't make the rest
		// of the pixels visited.
		while(prevSize!=unvisited.size()){
			prevSize=unvisited.size();
			for(int i=0; i<unvisited.size(); i++){
				xy=(XYPair)unvisited.elementAt(i);
				// if it's on the selection path skip it.
				if(visited[((XYPair)unvisited.get(i)).y][((XYPair)unvisited.get(i)).x]==2)
					continue;
				
				// Check if any of the neighbors are now visited.
				if((visited[xy.y][xy.x]=checkHeighbors(visited, xy.x, xy.y))==1){
					// it was visited so remove it from the list.
					unvisited.remove(i);
					i--;
				}
			}
		}

		// Create the matte from the result above.
		BufferedImage matte=new BufferedImage(maxX-minX+1, maxY-minY+1, BufferedImage.TYPE_INT_ARGB);
		for(int i=0; i<matte.getHeight(); i++){
			for(int j=0; j<matte.getWidth(); j++){
				if(visited[i+1][j+1]!=1)
					matte.setRGB(j, i, im.getRGB(j+minX, i+minY));
				else
					matte.setRGB(j, i, 0);
			}
		}
		return matte;
	}
}
